# ruff: noqa: UP006 UP007 UP045
import errno
import os
import typing as ta

from .ostypes import Fd


class PipeFds(ta.NamedTuple):
    r: Fd
    w: Fd


def make_pipe() -> PipeFds:
    return PipeFds(*os.pipe())  # type: ignore


def read_fd(fd: Fd) -> bytes:
    try:
        data = os.read(fd, 2 << 16)  # 128K
    except OSError as why:
        if why.args[0] not in (errno.EWOULDBLOCK, errno.EBADF, errno.EINTR):
            raise
        data = b''
    return data


def close_fd(fd: Fd) -> bool:
    try:
        os.close(fd)
    except OSError:
        return False
    return True


def is_fd_open(fd: Fd) -> bool:
    try:
        n = os.dup(fd)
    except OSError:
        return False
    os.close(n)
    return True


def get_open_fds(limit: int) -> ta.FrozenSet[Fd]:
    return frozenset(fd for i in range(limit) if is_fd_open(fd := Fd(i)))



# ruff: noqa: UP006 UP007 UP045
import dataclasses as dc
import typing as ta

from .platforms import Platform


##


@dc.dataclass(frozen=True)
class SystemConfig:
    platform: ta.Optional[Platform] = None



import abc
import typing as ta

from omlish import cached
from omlish import check
from omlish import collections as col
from omlish import dataclasses as dc
from omlish import lang


DateTime = ta.NewType('DateTime', str)
MillisecondDateTime = ta.NewType('MillisecondDateTime', str)

Timestamp = ta.NewType('Timestamp', str)


##


@dc.dataclass(frozen=True)
class Tag:
    key: str
    value: str


TagList: ta.TypeAlias = ta.Sequence[Tag]


##


class ValueType(abc.ABC):  # noqa
    pass


@dc.dataclass(frozen=True)
class ListValueType(ValueType):
    e: type


@dc.dataclass(frozen=True)
class MapValueType(ValueType):
    k: type
    v: type


##


class SHAPE_NAME(lang.Marker):  # noqa
    pass


def common_metadata(
        *,
        shape_name: str | None = None,
) -> dict[ta.Any, ta.Any]:
    md = {}

    if shape_name is not None:
        md[SHAPE_NAME] = shape_name

    return md


##


def shape_metadata(
        **kwargs: ta.Any,
) -> dict[ta.Any, ta.Any]:
    md = {**common_metadata(**kwargs)}

    return md


class ShapeInfo:
    def __init__(
            self,
            cls: type['Shape'],
            metadata: ta.Mapping[ta.Any, ta.Any],
    ) -> None:
        super().__init__()

        self._cls = check.issubclass(cls, Shape)
        self._metadata = metadata

    @property
    def cls(self) -> type['Shape']:
        return self._cls

    @property
    def metadata(self) -> ta.Mapping[ta.Any, ta.Any]:
        return self._metadata

    #

    @cached.property
    def fields(self) -> ta.Sequence[dc.Field]:
        check.state(dc.is_immediate_dataclass(self._cls))
        fls = dc.fields(self._cls)
        return fls  # noqa

    @cached.property
    def fields_by_name(self) -> ta.Mapping[str, dc.Field]:
        return col.make_map_by(lambda fl: fl.name, self.fields, strict=True)

    @cached.property
    def fields_by_member_name(self) -> ta.Mapping[str, dc.Field]:
        return col.make_map(
            [(n, f) for f in self.fields if (n := f.metadata.get(MEMBER_NAME)) is not None],
            strict=True,
        )

    @cached.property
    def fields_by_serialization_name(self) -> ta.Mapping[str, dc.Field]:
        l = []
        for f in self.fields:
            if sn := f.metadata.get(SERIALIZATION_NAME):
                l.append((sn, f))
            elif mn := f.metadata.get(MEMBER_NAME):
                l.append((mn, f))
        return col.make_map(l, strict=True)


@dc.dataclass(frozen=True, kw_only=True)
class Shape:
    __shape__: ta.ClassVar[ShapeInfo]

    def __init_subclass__(
            cls,
            *,
            shape_name: str | None = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init_subclass__(**kwargs)

        check.state(not hasattr(cls, '__shape__'))

        info = ShapeInfo(
            cls,
            shape_metadata(**kwargs),
        )

        cls.__shape__ = info


##


class MEMBER_NAME(lang.Marker):  # noqa
    pass


class SERIALIZATION_NAME(lang.Marker):  # noqa
    pass


class VALUE_TYPE(lang.Marker):  # noqa
    pass


#


def field_metadata(
        *,
        member_name: str | None = None,
        serialization_name: str | None = None,
        value_type: ValueType | None = None,
        **kwargs: ta.Any,
) -> dict[ta.Any, ta.Any]:
    md = {**common_metadata(**kwargs)}

    if member_name is not None:
        md[MEMBER_NAME] = member_name
    if serialization_name is not None:
        md[SERIALIZATION_NAME] = serialization_name
    if value_type is not None:
        md[VALUE_TYPE] = value_type

    return md


##


@dc.dataclass(frozen=True, eq=False, kw_only=True)
class Operation:
    name: str

    input: type[Shape] | None = None
    output: type[Shape] | None = None

    errors: ta.Sequence[type[Shape]] | None = None



# ruff: noqa: UP006 UP007 UP045
import dataclasses as dc
import typing as ta

from omlish.lite.check import check
from omlish.lite.logs import log

from ..commands.base import Command
from ..commands.base import CommandExecutor
from .packages import SystemPackage
from .packages import SystemPackageManager


##


@dc.dataclass(frozen=True)
class CheckSystemPackageCommand(Command['CheckSystemPackageCommand.Output']):
    pkgs: ta.Sequence[str] = ()

    def __post_init__(self) -> None:
        check.not_isinstance(self.pkgs, str)

    @dc.dataclass(frozen=True)
    class Output(Command.Output):
        pkgs: ta.Sequence[SystemPackage]


class CheckSystemPackageCommandExecutor(CommandExecutor[CheckSystemPackageCommand, CheckSystemPackageCommand.Output]):
    def __init__(
            self,
            *,
            mgr: SystemPackageManager,
    ) -> None:
        super().__init__()

        self._mgr = mgr

    async def execute(self, cmd: CheckSystemPackageCommand) -> CheckSystemPackageCommand.Output:
        log.info('Checking system package!')

        ret = await self._mgr.query(*cmd.pkgs)

        return CheckSystemPackageCommand.Output(list(ret.values()))



# ruff: noqa: UP006 UP007 UP045
import grp
import os
import pwd
import typing as ta


def drop_privileges(user: ta.Union[int, str, None]) -> ta.Optional[str]:
    """
    Drop privileges to become the specified user, which may be a username or uid. Called for supervisord startup and
    when spawning subprocesses. Returns None on success or a string error message if privileges could not be dropped.
    """

    if user is None:
        return 'No user specified to setuid to!'

    # get uid for user, which can be a number or username
    try:
        uid = int(user)
    except ValueError:
        try:
            pwrec = pwd.getpwnam(user)  # type: ignore
        except KeyError:
            return f"Can't find username {user!r}"
        uid = pwrec[2]
    else:
        try:
            pwrec = pwd.getpwuid(uid)
        except KeyError:
            return f"Can't find uid {uid!r}"

    current_uid = os.getuid()

    if current_uid == uid:
        # do nothing and return successfully if the uid is already the current one. this allows a supervisord running as
        # an unprivileged user "foo" to start a process where the config has "user=foo" (same user) in it.
        return None

    if current_uid != 0:
        return "Can't drop privilege as nonroot user"

    gid = pwrec[3]
    if hasattr(os, 'setgroups'):
        user = pwrec[0]
        groups = [grprec[2] for grprec in grp.getgrall() if user in grprec[3]]

        # always put our primary gid first in this list, otherwise we can lose group info since sometimes the first
        # group in the setgroups list gets overwritten on the subsequent setgid call (at least on freebsd 9 with
        # python 2.7 - this will be safe though for all unix /python version combos)
        groups.insert(0, gid)
        try:
            os.setgroups(groups)
        except OSError:
            return 'Could not set groups of effective user'

    try:
        os.setgid(gid)
    except OSError:
        return 'Could not set group id of effective user'

    os.setuid(uid)

    return None
